/*
 * State.java
 *
 * Created on 21. listopad 2006, 20:17
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package osmdam;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.TreeMap;

/**
 *
 * @author pirko
 */
public class State {
    
    private Map<Character,Dama> damy = new TreeMap<Character,Dama>();
    
    private static final Character[] cols = {'a','b','c','d','e','f','g','h'};
    
    public enum Type {
        DEFAULT,
        SOFT,
        HARD,
        GEN
    }
    
    /** Creates a new instance of State */
    public State(Type type) {
        switch (type) {
            case DEFAULT:
                initFill();
                break;
            case SOFT:
                initFillSoft();
                break;
            case HARD:
                initFillHard();
                break;
            case GEN:
                initFillGen();
                break;
        }
        recalcConflicts();        
    }
    
    public State(Map<Character,Dama> damy) {
        this.setDamy(damy);
        recalcConflicts();
    }
    
    private void initFill() {
        Dama d1 = new Dama('a',4,3);
        Dama d2 = new Dama('b',3,5);
        Dama d3 = new Dama('c',2,4);
        Dama d4 = new Dama('d',5,4);
        Dama d5 = new Dama('e',4,5);
        Dama d6 = new Dama('f',3,5);
        Dama d7 = new Dama('g',2,5);
        Dama d8 = new Dama('h',3,3);
        
        getDamy().put(d1.getCol(),d1);
        getDamy().put(d2.getCol(),d2);
        getDamy().put(d3.getCol(),d3);
        getDamy().put(d4.getCol(),d4);
        getDamy().put(d5.getCol(),d5);
        getDamy().put(d6.getCol(),d6);
        getDamy().put(d7.getCol(),d7);
        getDamy().put(d8.getCol(),d8);
    }
    
    private void initFillHard() {
        Dama d1 = new Dama('a',1,0);
        Dama d2 = new Dama('b',3,0);
        Dama d3 = new Dama('c',4,0);
        Dama d4 = new Dama('d',6,0);
        Dama d5 = new Dama('e',4,0);
        Dama d6 = new Dama('f',2,0);
        Dama d7 = new Dama('g',5,0);
        Dama d8 = new Dama('h',7,0);
        
        getDamy().put(d1.getCol(),d1);
        getDamy().put(d2.getCol(),d2);
        getDamy().put(d3.getCol(),d3);
        getDamy().put(d4.getCol(),d4);
        getDamy().put(d5.getCol(),d5);
        getDamy().put(d6.getCol(),d6);
        getDamy().put(d7.getCol(),d7);
        getDamy().put(d8.getCol(),d8);
    }
    
    private void initFillSoft() {
        Dama d1 = new Dama('a',7,0);
        Dama d2 = new Dama('b',4,0);
        Dama d3 = new Dama('c',2,0);
        Dama d4 = new Dama('d',5,1);
        Dama d5 = new Dama('e',8,0);
        Dama d6 = new Dama('f',6,0);
        Dama d7 = new Dama('g',3,0);
        Dama d8 = new Dama('h',1,1);
        
        getDamy().put(d1.getCol(),d1);
        getDamy().put(d2.getCol(),d2);
        getDamy().put(d3.getCol(),d3);
        getDamy().put(d4.getCol(),d4);
        getDamy().put(d5.getCol(),d5);
        getDamy().put(d6.getCol(),d6);
        getDamy().put(d7.getCol(),d7);
        getDamy().put(d8.getCol(),d8);
    }
    
    private void initFillGen() {
        Random r = new Random();
        Dama d1 = new Dama('a',r.nextInt(7)+1,0);
        Dama d2 = new Dama('b',r.nextInt(7)+1,0);
        Dama d3 = new Dama('c',r.nextInt(7)+1,0);
        Dama d4 = new Dama('d',r.nextInt(7)+1,1);
        Dama d5 = new Dama('e',r.nextInt(7)+1,0);
        Dama d6 = new Dama('f',r.nextInt(7)+1,0);
        Dama d7 = new Dama('g',r.nextInt(7)+1,0);
        Dama d8 = new Dama('h',r.nextInt(7)+1,1);
        
        getDamy().put(d1.getCol(),d1);
        getDamy().put(d2.getCol(),d2);
        getDamy().put(d3.getCol(),d3);
        getDamy().put(d4.getCol(),d4);
        getDamy().put(d5.getCol(),d5);
        getDamy().put(d6.getCol(),d6);
        getDamy().put(d7.getCol(),d7);
        getDamy().put(d8.getCol(),d8);
    }
    
   
   // TODO prepsat na generovani svech 56 nasledniku 
    public List<State> getNext() {
        int best = value();         
        List<State> res = new ArrayList<State>();
        
        for (Dama dama : getDamy().values()) {
            for (int i=1;i<=8;i++) {
                if (i == dama.getRow())
                    continue;
                
                Map<Character,Dama> newDamy = new TreeMap<Character,Dama>();
                for (Character c : cols) {
                    Dama d = getDamy().get(c);
                    newDamy.put(d.getCol(),new Dama(d.getCol(),d.getRow(),d.getConflict()));
                }
                newDamy.put(dama.getCol(), new Dama(dama.getCol(),i,dama.getConflict()));
                
                State state = new State(newDamy);
                
                int val = state.value();
                if (val < best) {
                    best = val;
                    res.clear();
                    res.add(state);
                } else if (val == best) {
                    res.add(state);
                }
            }
        }
        return res;
    }
    
    private List<Dama> getConflicting() {
        List<Dama> res = new ArrayList<Dama>();
        for (Character c : cols) {
            Dama d = getDamy().get(c);
            if (d.getConflict() > 0)
                res.add(d);
        }
        return res;
    }
    
    private int indexOfCol(char col) {
        return Arrays.binarySearch(cols,col);
    }
    
    private int calcConflict(char col,int row) {
        int res = 0;
        
        // kontrolujem radek
        for (Character c : cols) {
            if (c.charValue() != col && getDamy().get(c).getRow() == row)
                res++;
        }
        
        int colIndex = indexOfCol(col);
        // leva diagonala
        int r = row + 1;
        int c = colIndex -1;
        while (r<=8 && c>=0) {
            if (getDamy().get(cols[c]).getRow() == r)
                res++;
            r++;
            c--;
        }
        r = row - 1;
        c = colIndex + 1;
        while (r>0 && c<8) {
            if (getDamy().get(cols[c]).getRow() == r)
                res++;
            r--;
            c++;
        }
        
        // prava diagonala
        r = row + 1;
        c = colIndex + 1;
        while (r<=8 && c<8) {
            if (getDamy().get(cols[c]).getRow() == r)
                res++;
            r++;
            c++;
        }
        r = row - 1;
        c = colIndex - 1;
        while (r>0 && c>=0) {
            if (getDamy().get(cols[c]).getRow() == r)
                res++;
            r--;
            c--;
        }
        
        return res;
    }
    
    private void recalcConflicts() {
        for (Character c : cols) {
            Dama d = getDamy().get(c);
            d.setConflict(calcConflict(d.getCol(),d.getRow()));
        }
    }
    
    public void print() {
        for (Character c : cols) {
            System.out.print(getDamy().get(c)+",");            
        }
        System.out.print("\n");
    }
    
    public boolean equals(Object o) {
        if (o instanceof State) {
            State s = (State)o;
            boolean res = true;
            for (Character c : cols) {
                res = res && damy.get(c).equals(s.getDamy().get(c));
            }
            return res;
        } else
            return false;
    }

    public Map<Character, Dama> getDamy() {
        return damy;
    }

    public void setDamy(Map<Character, Dama> damy) {
        this.damy = damy;
    }
    
    public boolean isSolution() {
        return getConflicting().size() == 0;
    }
    
    public int value() {
        int res = 0;
        for (Dama d : getDamy().values()) {
            res += d.getConflict();
        }
        return res/2;
    }
}
